"""Module containing extension app authenticator classes.
"""
from abc import ABC, abstractmethod

from symphony.bdk.core.auth.auth_session import AppAuthSession
from symphony.bdk.core.auth.jwt_helper import validate_jwt, create_signed_jwt
from symphony.bdk.core.auth.tokens_repository import TokensRepository, InMemoryTokensRepository
from symphony.bdk.core.config.model.bdk_retry_config import BdkRetryConfig
from symphony.bdk.core.config.model.bdk_rsa_key_config import BdkRsaKeyConfig
from symphony.bdk.core.retry import retry
from symphony.bdk.core.retry.strategy import authentication_retry
from symphony.bdk.gen.auth_api.certificate_authentication_api import CertificateAuthenticationApi
from symphony.bdk.gen.auth_api.certificate_pod_api import CertificatePodApi
from symphony.bdk.gen.auth_model.extension_app_authenticate_request import ExtensionAppAuthenticateRequest
from symphony.bdk.gen.login_api.authentication_api import AuthenticationApi
from symphony.bdk.gen.login_model.authenticate_extension_app_request import AuthenticateExtensionAppRequest
from symphony.bdk.gen.login_model.extension_app_tokens import ExtensionAppTokens
from symphony.bdk.gen.pod_api.pod_api import PodApi
from symphony.bdk.gen.pod_model.pod_certificate import PodCertificate


class ExtensionAppAuthenticator(ABC):
    """Base abstract class to handle extension app authentication.
    """

    def __init__(self, app_id: str, tokens_repository: TokensRepository = None):
        """

        :param app_id: the application ID
        :param tokens_repository: the tokens repository to store existing valid sessions.
          Defaults to InMemoryTokensRepository
        """
        self._app_id = app_id
        self._tokens_repository = tokens_repository or InMemoryTokensRepository()

    async def authenticate_extension_app(self, app_token: str) -> AppAuthSession:
        """Authenticates an extension app.

        :param app_token: the token generated by the application backend
        :return: the extension app authentication session
        """
        auth_session = AppAuthSession(self, app_token)
        await auth_session.refresh()

        return auth_session

    async def validate_jwt(self, jwt: str) -> dict:
        """Validates a jwt against the pod certificate.

        :param jwt: the jwt to be validated
        :return: the dictionary of jwt claims
        :raise AuthInitializationError: If the pod certificate or jwt are invalid.
        """
        pod_certificate = await self._get_pod_certificate()
        return validate_jwt(jwt, pod_certificate.certificate, self._app_id)

    async def authenticate_and_retrieve_tokens(self, app_token: str) -> ExtensionAppTokens:
        """Actually authenticates and retrieves the tokens.
        Save the tokens in the tokens repository.

        :param app_token: the token generated by the application backend
        :return: the extension app tokens
        """
        ext_app_tokens = await self._retrieve_tokens(app_token)
        await self._tokens_repository.save(ext_app_tokens.app_token, ext_app_tokens.symphony_token)

        return ext_app_tokens

    async def is_token_pair_valid(self, app_token: str, symphony_token: str) -> bool:
        """Validates if appToken and symphonyToken corresponds to an existing session.

        :param app_token: the application token
        :param symphony_token: the Symphony token
        :return: True if token pair is valid, False otherwise
        """
        retrieved_sym_token = await self._tokens_repository.get(app_token)
        return retrieved_sym_token is not None and retrieved_sym_token == symphony_token

    @abstractmethod
    async def _get_pod_certificate(self) -> PodCertificate:
        """Retrieves the pod certificate. Used when validating JWT tokens.

        :return: the pod certificate
        """

    @abstractmethod
    async def _retrieve_tokens(self, app_token: str) -> ExtensionAppTokens:
        """Actually authenticates and retrieves the tokens.

        :param app_token: the token generated by the application backend
        :return: the extension app tokens
        """


class ExtensionAppAuthenticatorRsa(ExtensionAppAuthenticator):
    """A subclass of :class:`ExtensionAppAuthenticator` specific to extension app RSA authentication.
    """

    def __init__(self,
                 authentication_api: AuthenticationApi,
                 pod_api: PodApi,
                 app_id: str,
                 private_key_config: BdkRsaKeyConfig,
                 retry_config: BdkRetryConfig,
                 tokens_repository: TokensRepository = None):
        """

        :param authentication_api: the AuthenticationApi instance
        :param pod_api: the PodApi instance
        :param app_id: the application ID
        :param private_key_config: the private key configuration of the extension app
        :param retry_config: retry configuration
        :param tokens_repository: the tokens repository to store existing valid sessions.
          Defaults to InMemoryTokensRepository
        """
        super().__init__(app_id, tokens_repository)
        self._authentication_api = authentication_api
        self._pod_api = pod_api
        self._private_key_config = private_key_config
        self._retry_config = retry_config

    @retry(retry=authentication_retry)
    async def _retrieve_tokens(self, app_token: str) -> ExtensionAppTokens:
        jwt = create_signed_jwt(self._private_key_config, self._app_id)
        authentication_request = AuthenticateExtensionAppRequest(app_token=app_token, auth_token=jwt)

        return await self._authentication_api.v1_pubkey_app_authenticate_extension_app_post(authentication_request)

    @retry(retry=authentication_retry)
    async def _get_pod_certificate(self) -> PodCertificate:
        return await self._pod_api.v1_podcert_get()


class ExtensionAppAuthenticatorCert(ExtensionAppAuthenticator):
    """A subclass of :class:`ExtensionAppAuthenticator` specific to extension app certificate authentication.
    """

    def __init__(self,
                 certificate_authentication_api: CertificateAuthenticationApi,
                 certificate_pod_api: CertificatePodApi,
                 app_id: str,
                 retry_config: BdkRetryConfig,
                 tokens_repository: TokensRepository = None):
        """

        :param certificate_authentication_api: the CertificateAuthenticationApi instance
        :param certificate_pod_api: the CertificatePodApi instance
        :param app_id: the application ID
        :param retry_config: Retry configuration
        :param tokens_repository: the tokens repository to store existing valid sessions.
          Defaults to InMemoryTokensRepository
        """
        super().__init__(app_id, tokens_repository)
        self._certificate_authentication_api = certificate_authentication_api
        self._certificate_pod_api = certificate_pod_api
        self._retry_config = retry_config

    @retry(retry=authentication_retry)
    async def _retrieve_tokens(self, app_token: str) -> ExtensionAppTokens:
        authentication_request = ExtensionAppAuthenticateRequest(app_token=app_token)
        return await self._certificate_authentication_api.v1_authenticate_extension_app_post(
            auth_request=authentication_request)

    @retry(retry=authentication_retry)
    async def _get_pod_certificate(self) -> PodCertificate:
        return await self._certificate_pod_api.v1_app_pod_certificate_get()



"""Module containing BotAuthenticator classes.
"""
from abc import ABC, abstractmethod

from symphony.bdk.core.auth.jwt_helper import create_signed_jwt
from symphony.bdk.core.config.model.bdk_bot_config import BdkBotConfig
from symphony.bdk.core.config.model.bdk_retry_config import BdkRetryConfig
from symphony.bdk.core.retry import retry
from symphony.bdk.core.retry.strategy import authentication_retry
from symphony.bdk.gen.api_client import ApiClient
from symphony.bdk.gen.auth_api.certificate_authentication_api import CertificateAuthenticationApi
from symphony.bdk.gen.login_api.authentication_api import AuthenticationApi
from symphony.bdk.gen.login_model.authenticate_request import AuthenticateRequest


class BotAuthenticator(ABC):
    """Bot authentication service.
    """

    def __init__(self, session_auth_client: ApiClient, key_manager_auth_client: ApiClient,
                 retry_config: BdkRetryConfig):
        self._session_auth_client = session_auth_client
        self._key_manager_auth_client = key_manager_auth_client
        self._retry_config = retry_config

    async def retrieve_session_token(self) -> str:
        """Authenticates and retrieves a new session token.

        :return: the retrieved session token.
        """
        return await self._authenticate_and_get_token(self._session_auth_client)

    async def retrieve_key_manager_token(self) -> str:
        """Authenticated and retrieved a new key manager session.

        :return: the retrieved key manager session.
        """
        return await self._authenticate_and_get_token(self._key_manager_auth_client)

    @abstractmethod
    async def _authenticate_and_get_token(self, api_client: ApiClient) -> str:
        """

        :param api_client: the api client instance to use to retrieve the token
          (either self._session_auth_client or self._key_manager_auth_client)
        :return: the token as a string
        """


class BotAuthenticatorRsa(BotAuthenticator):
    """Bot authenticator RSA implementation.
    """

    def __init__(self, bot_config: BdkBotConfig, login_api_client: ApiClient, relay_api_client: ApiClient,
                 retry_config: BdkRetryConfig):
        self._bot_config = bot_config
        super().__init__(login_api_client, relay_api_client, retry_config)

    @retry(retry=authentication_retry)
    async def _authenticate_and_get_token(self, api_client: ApiClient) -> str:
        jwt = create_signed_jwt(self._bot_config.private_key, self._bot_config.username)
        req = AuthenticateRequest(token=jwt)

        token = await AuthenticationApi(api_client).pubkey_authenticate_post(req)
        return token.token


class BotAuthenticatorCert(BotAuthenticator):
    """Bot authenticator certificate implementation.
    """

    @retry(retry=authentication_retry)
    async def _authenticate_and_get_token(self, api_client: ApiClient) -> str:
        token = await CertificateAuthenticationApi(api_client).v1_authenticate_post()
        return token.token



"""
    Pod API

    This document refers to Symphony API calls that do not need encryption or decryption of content.  - sessionToken can be obtained by calling the authenticationAPI on the symphony back end and the key manager respectively. Refer to the methods described in authenticatorAPI.yaml. - Actions are defined to be atomic, ie will succeed in their entirety or fail and have changed nothing. - If it returns a 40X status then it will have made no change to the system even if ome subset of the request would have succeeded. - If this contract cannot be met for any reason then this is an error and the response code will be 50X.   # noqa: E501

    The version of the OpenAPI document: 20.12.1
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from symphony.bdk.gen.api_client import ApiClient, Endpoint as _Endpoint
from symphony.bdk.gen.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from symphony.bdk.gen.pod_model.avatar_list import AvatarList
from symphony.bdk.gen.pod_model.avatar_update import AvatarUpdate
from symphony.bdk.gen.pod_model.delegate_action import DelegateAction
from symphony.bdk.gen.pod_model.disclaimer import Disclaimer
from symphony.bdk.gen.pod_model.error import Error
from symphony.bdk.gen.pod_model.feature_list import FeatureList
from symphony.bdk.gen.pod_model.followers_list import FollowersList
from symphony.bdk.gen.pod_model.followers_list_response import FollowersListResponse
from symphony.bdk.gen.pod_model.following_list_response import FollowingListResponse
from symphony.bdk.gen.pod_model.integer_list import IntegerList
from symphony.bdk.gen.pod_model.role_detail_list import RoleDetailList
from symphony.bdk.gen.pod_model.string_id import StringId
from symphony.bdk.gen.pod_model.success_response import SuccessResponse
from symphony.bdk.gen.pod_model.user_detail_list import UserDetailList
from symphony.bdk.gen.pod_model.user_filter import UserFilter
from symphony.bdk.gen.pod_model.user_status import UserStatus
from symphony.bdk.gen.pod_model.user_suspension import UserSuspension
from symphony.bdk.gen.pod_model.v2_user_attributes import V2UserAttributes
from symphony.bdk.gen.pod_model.v2_user_create import V2UserCreate
from symphony.bdk.gen.pod_model.v2_user_detail import V2UserDetail
from symphony.bdk.gen.pod_model.v2_user_detail_list import V2UserDetailList


class UserApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __v1_admin_system_roles_list_get(
            self,
            session_token,
            **kwargs
        ):
            """Get a list of all roles available in the company (pod)  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_system_roles_list_get(session_token, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                RoleDetailList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            return self.call_with_http_info(**kwargs)

        self.v1_admin_system_roles_list_get = _Endpoint(
            settings={
                'response_type': (RoleDetailList,),
                'auth': [],
                'endpoint_path': '/v1/admin/system/roles/list',
                'operation_id': 'v1_admin_system_roles_list_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                ],
                'required': [
                    'session_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                },
                'location_map': {
                    'session_token': 'header',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_admin_system_roles_list_get
        )

        def __v1_admin_user_find_post(
            self,
            session_token,
            payload,
            **kwargs
        ):
            """Find a user based on attributes  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_find_post(session_token, payload, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                payload (UserFilter):

            Keyword Args:
                skip (int): Number of users to skip. Default is 0. Must be a integer equals or bigger than 0.. [optional]
                limit (int): Maximum number of users to return. Default is 100. Must be a positive integer and must not exceed 1000.. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                UserDetailList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['payload'] = \
                payload
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_find_post = _Endpoint(
            settings={
                'response_type': (UserDetailList,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/find',
                'operation_id': 'v1_admin_user_find_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'payload',
                    'skip',
                    'limit',
                ],
                'required': [
                    'session_token',
                    'payload',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'payload':
                        (UserFilter,),
                    'skip':
                        (int,),
                    'limit':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'skip': 'skip',
                    'limit': 'limit',
                },
                'location_map': {
                    'session_token': 'header',
                    'payload': 'body',
                    'skip': 'query',
                    'limit': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_admin_user_find_post
        )

        def __v1_admin_user_uid_avatar_get(
            self,
            session_token,
            uid,
            **kwargs
        ):
            """Get the URL of the avatar of a particular user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_avatar_get(session_token, uid, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                AvatarList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_avatar_get = _Endpoint(
            settings={
                'response_type': (AvatarList,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/avatar',
                'operation_id': 'v1_admin_user_uid_avatar_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                ],
                'required': [
                    'session_token',
                    'uid',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_avatar_get
        )

        def __v1_admin_user_uid_avatar_update_post(
            self,
            session_token,
            uid,
            payload,
            **kwargs
        ):
            """Update the avatar of a particular user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_avatar_update_post(session_token, uid, payload, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 
                payload (AvatarUpdate):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            kwargs['payload'] = \
                payload
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_avatar_update_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/avatar/update',
                'operation_id': 'v1_admin_user_uid_avatar_update_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'required': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'payload':
                        (AvatarUpdate,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'payload': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_avatar_update_post
        )

        def __v1_admin_user_uid_delegates_get(
            self,
            session_token,
            uid,
            **kwargs
        ):
            """Get the delegates assigned to a user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_delegates_get(session_token, uid, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                IntegerList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_delegates_get = _Endpoint(
            settings={
                'response_type': (IntegerList,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/delegates',
                'operation_id': 'v1_admin_user_uid_delegates_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                ],
                'required': [
                    'session_token',
                    'uid',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_delegates_get
        )

        def __v1_admin_user_uid_delegates_update_post(
            self,
            session_token,
            uid,
            payload,
            **kwargs
        ):
            """Update the delegates assigned to a user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_delegates_update_post(session_token, uid, payload, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 
                payload (DelegateAction):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            kwargs['payload'] = \
                payload
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_delegates_update_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/delegates/update',
                'operation_id': 'v1_admin_user_uid_delegates_update_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'required': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'payload':
                        (DelegateAction,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'payload': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_delegates_update_post
        )

        def __v1_admin_user_uid_disclaimer_delete(
            self,
            session_token,
            uid,
            **kwargs
        ):
            """Unassign a disclaimer from a user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_disclaimer_delete(session_token, uid, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_disclaimer_delete = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/disclaimer',
                'operation_id': 'v1_admin_user_uid_disclaimer_delete',
                'http_method': 'DELETE',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                ],
                'required': [
                    'session_token',
                    'uid',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_disclaimer_delete
        )

        def __v1_admin_user_uid_disclaimer_get(
            self,
            session_token,
            uid,
            **kwargs
        ):
            """Get the disclaimer assigned to a user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_disclaimer_get(session_token, uid, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Disclaimer
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_disclaimer_get = _Endpoint(
            settings={
                'response_type': (Disclaimer,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/disclaimer',
                'operation_id': 'v1_admin_user_uid_disclaimer_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                ],
                'required': [
                    'session_token',
                    'uid',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_disclaimer_get
        )

        def __v1_admin_user_uid_disclaimer_update_post(
            self,
            session_token,
            uid,
            payload,
            **kwargs
        ):
            """Assign a disclaimer to a user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_disclaimer_update_post(session_token, uid, payload, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 
                payload (StringId):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            kwargs['payload'] = \
                payload
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_disclaimer_update_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/disclaimer/update',
                'operation_id': 'v1_admin_user_uid_disclaimer_update_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'required': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'payload':
                        (StringId,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'payload': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_disclaimer_update_post
        )

        def __v1_admin_user_uid_features_get(
            self,
            session_token,
            uid,
            **kwargs
        ):
            """Get the list of Symphony feature entitlements enabled for a particular user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_features_get(session_token, uid, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                FeatureList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_features_get = _Endpoint(
            settings={
                'response_type': (FeatureList,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/features',
                'operation_id': 'v1_admin_user_uid_features_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                ],
                'required': [
                    'session_token',
                    'uid',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_features_get
        )

        def __v1_admin_user_uid_features_update_post(
            self,
            session_token,
            uid,
            payload,
            **kwargs
        ):
            """Update the list of Symphony feature entitlements for a particular user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_features_update_post(session_token, uid, payload, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 
                payload (FeatureList):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            kwargs['payload'] = \
                payload
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_features_update_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/features/update',
                'operation_id': 'v1_admin_user_uid_features_update_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'required': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'payload':
                        (FeatureList,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'payload': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_features_update_post
        )

        def __v1_admin_user_uid_roles_add_post(
            self,
            session_token,
            uid,
            payload,
            **kwargs
        ):
            """Add a role to a user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_roles_add_post(session_token, uid, payload, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 
                payload (StringId):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            kwargs['payload'] = \
                payload
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_roles_add_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/roles/add',
                'operation_id': 'v1_admin_user_uid_roles_add_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'required': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'payload':
                        (StringId,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'payload': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_roles_add_post
        )

        def __v1_admin_user_uid_roles_remove_post(
            self,
            session_token,
            uid,
            payload,
            **kwargs
        ):
            """Remove a role from a user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_roles_remove_post(session_token, uid, payload, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 
                payload (StringId):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            kwargs['payload'] = \
                payload
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_roles_remove_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/roles/remove',
                'operation_id': 'v1_admin_user_uid_roles_remove_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'required': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'payload':
                        (StringId,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'payload': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_roles_remove_post
        )

        def __v1_admin_user_uid_status_get(
            self,
            session_token,
            uid,
            **kwargs
        ):
            """Get the status, active or inactive, for a particular user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_status_get(session_token, uid, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                UserStatus
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_status_get = _Endpoint(
            settings={
                'response_type': (UserStatus,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/status',
                'operation_id': 'v1_admin_user_uid_status_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                ],
                'required': [
                    'session_token',
                    'uid',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_status_get
        )

        def __v1_admin_user_uid_status_update_post(
            self,
            session_token,
            uid,
            payload,
            **kwargs
        ):
            """Update the status of a particular user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_uid_status_update_post(session_token, uid, payload, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 
                payload (UserStatus):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            kwargs['payload'] = \
                payload
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_uid_status_update_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{uid}/status/update',
                'operation_id': 'v1_admin_user_uid_status_update_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'required': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'payload':
                        (UserStatus,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'payload': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_admin_user_uid_status_update_post
        )

        def __v1_admin_user_user_id_suspension_update_put(
            self,
            session_token,
            user_id,
            payload,
            **kwargs
        ):
            """Update the status of suspension of a particular user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_admin_user_user_id_suspension_update_put(session_token, user_id, payload, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token
                user_id (int): User ID as a decimal integer
                payload (UserSuspension):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['user_id'] = \
                user_id
            kwargs['payload'] = \
                payload
            return self.call_with_http_info(**kwargs)

        self.v1_admin_user_user_id_suspension_update_put = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/admin/user/{userId}/suspension/update',
                'operation_id': 'v1_admin_user_user_id_suspension_update_put',
                'http_method': 'PUT',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'user_id',
                    'payload',
                ],
                'required': [
                    'session_token',
                    'user_id',
                    'payload',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'user_id':
                        (int,),
                    'payload':
                        (UserSuspension,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'user_id': 'userId',
                },
                'location_map': {
                    'session_token': 'header',
                    'user_id': 'path',
                    'payload': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_admin_user_user_id_suspension_update_put
        )

        def __v1_user_uid_follow_post(
            self,
            session_token,
            uid,
            uid_list,
            **kwargs
        ):
            """Make a list of users start following a specific user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_user_uid_follow_post(session_token, uid, uid_list, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer of the user to be followed
                uid_list (FollowersList): List of (integer) User IDs of the followers

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            kwargs['uid_list'] = \
                uid_list
            return self.call_with_http_info(**kwargs)

        self.v1_user_uid_follow_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/user/{uid}/follow',
                'operation_id': 'v1_user_uid_follow_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'uid_list',
                ],
                'required': [
                    'session_token',
                    'uid',
                    'uid_list',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'uid_list':
                        (FollowersList,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'uid_list': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_user_uid_follow_post
        )

        def __v1_user_uid_followers_get(
            self,
            session_token,
            uid,
            **kwargs
        ):
            """Returns the list of followers for a specific user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_user_uid_followers_get(session_token, uid, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer of the user we want to get the followers list

            Keyword Args:
                limit (int): This is the maximum number of objects that may be returned.. [optional]
                before (str): Returns results from an opaque “before” cursor value as presented via a response cursor.. [optional]
                after (str): Returns results from an opaque “after” cursor value as presented via a response cursor.. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                FollowersListResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            return self.call_with_http_info(**kwargs)

        self.v1_user_uid_followers_get = _Endpoint(
            settings={
                'response_type': (FollowersListResponse,),
                'auth': [],
                'endpoint_path': '/v1/user/{uid}/followers',
                'operation_id': 'v1_user_uid_followers_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'limit',
                    'before',
                    'after',
                ],
                'required': [
                    'session_token',
                    'uid',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'limit':
                        (int,),
                    'before':
                        (str,),
                    'after':
                        (str,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                    'limit': 'limit',
                    'before': 'before',
                    'after': 'after',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'limit': 'query',
                    'before': 'query',
                    'after': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_user_uid_followers_get
        )

        def __v1_user_uid_following_get(
            self,
            session_token,
            uid,
            **kwargs
        ):
            """Returns the list of users that a specific user is following  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_user_uid_following_get(session_token, uid, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer of the user we want to get the following list

            Keyword Args:
                limit (int): This is the maximum number of objects that may be returned.. [optional]
                before (str): Returns results from an opaque “before” cursor value as presented via a response cursor.. [optional]
                after (str): Returns results from an opaque “after” cursor value as presented via a response cursor.. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                FollowingListResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            return self.call_with_http_info(**kwargs)

        self.v1_user_uid_following_get = _Endpoint(
            settings={
                'response_type': (FollowingListResponse,),
                'auth': [],
                'endpoint_path': '/v1/user/{uid}/following',
                'operation_id': 'v1_user_uid_following_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'limit',
                    'before',
                    'after',
                ],
                'required': [
                    'session_token',
                    'uid',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'limit':
                        (int,),
                    'before':
                        (str,),
                    'after':
                        (str,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                    'limit': 'limit',
                    'before': 'before',
                    'after': 'after',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'limit': 'query',
                    'before': 'query',
                    'after': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_user_uid_following_get
        )

        def __v1_user_uid_unfollow_post(
            self,
            session_token,
            uid,
            uid_list,
            **kwargs
        ):
            """Make a list of users unfollowing a specific user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_user_uid_unfollow_post(session_token, uid, uid_list, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer of the user to be unfollowed
                uid_list (FollowersList): List of (integer) User IDs of the followers

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            kwargs['uid_list'] = \
                uid_list
            return self.call_with_http_info(**kwargs)

        self.v1_user_uid_unfollow_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/user/{uid}/unfollow',
                'operation_id': 'v1_user_uid_unfollow_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'uid_list',
                ],
                'required': [
                    'session_token',
                    'uid',
                    'uid_list',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'uid_list':
                        (FollowersList,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'uid_list': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_user_uid_unfollow_post
        )

        def __v2_admin_user_create_post(
            self,
            session_token,
            payload,
            **kwargs
        ):
            """Create a new V2 User  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v2_admin_user_create_post(session_token, payload, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                payload (V2UserCreate):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                V2UserDetail
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['payload'] = \
                payload
            return self.call_with_http_info(**kwargs)

        self.v2_admin_user_create_post = _Endpoint(
            settings={
                'response_type': (V2UserDetail,),
                'auth': [],
                'endpoint_path': '/v2/admin/user/create',
                'operation_id': 'v2_admin_user_create_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'payload',
                ],
                'required': [
                    'session_token',
                    'payload',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'payload':
                        (V2UserCreate,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                },
                'location_map': {
                    'session_token': 'header',
                    'payload': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v2_admin_user_create_post
        )

        def __v2_admin_user_list_get(
            self,
            session_token,
            **kwargs
        ):
            """Retrieve a list of all users in the company (pod)  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v2_admin_user_list_get(session_token, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session Authentication token.

            Keyword Args:
                skip (int): Number of users to skip. Default is 0. Must be a integer equals or bigger than 0.. [optional]
                limit (int): Maximum number of users to return. Default is 100. Must be a positive integer and must not exceed 1000.. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                V2UserDetailList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            return self.call_with_http_info(**kwargs)

        self.v2_admin_user_list_get = _Endpoint(
            settings={
                'response_type': (V2UserDetailList,),
                'auth': [],
                'endpoint_path': '/v2/admin/user/list',
                'operation_id': 'v2_admin_user_list_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'skip',
                    'limit',
                ],
                'required': [
                    'session_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'skip':
                        (int,),
                    'limit':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'skip': 'skip',
                    'limit': 'limit',
                },
                'location_map': {
                    'session_token': 'header',
                    'skip': 'query',
                    'limit': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v2_admin_user_list_get
        )

        def __v2_admin_user_uid_get(
            self,
            session_token,
            uid,
            **kwargs
        ):
            """Retreive V2 User details for a particular user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v2_admin_user_uid_get(session_token, uid, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                V2UserDetail
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            return self.call_with_http_info(**kwargs)

        self.v2_admin_user_uid_get = _Endpoint(
            settings={
                'response_type': (V2UserDetail,),
                'auth': [],
                'endpoint_path': '/v2/admin/user/{uid}',
                'operation_id': 'v2_admin_user_uid_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                ],
                'required': [
                    'session_token',
                    'uid',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v2_admin_user_uid_get
        )

        def __v2_admin_user_uid_update_post(
            self,
            session_token,
            uid,
            payload,
            **kwargs
        ):
            """Update an existing V2 User  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v2_admin_user_uid_update_post(session_token, uid, payload, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                uid (int): User ID as a decimal integer 
                payload (V2UserAttributes):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                V2UserDetail
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['uid'] = \
                uid
            kwargs['payload'] = \
                payload
            return self.call_with_http_info(**kwargs)

        self.v2_admin_user_uid_update_post = _Endpoint(
            settings={
                'response_type': (V2UserDetail,),
                'auth': [],
                'endpoint_path': '/v2/admin/user/{uid}/update',
                'operation_id': 'v2_admin_user_uid_update_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'required': [
                    'session_token',
                    'uid',
                    'payload',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'uid':
                        (int,),
                    'payload':
                        (V2UserAttributes,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'uid': 'uid',
                },
                'location_map': {
                    'session_token': 'header',
                    'uid': 'path',
                    'payload': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v2_admin_user_uid_update_post
        )



from symphony.bdk.core.auth.auth_session import AuthSession
from symphony.bdk.core.config.model.bdk_retry_config import BdkRetryConfig
from symphony.bdk.core.retry import retry
from symphony.bdk.gen.pod_api.session_api import SessionApi
from symphony.bdk.gen.pod_model.user_v2 import UserV2


class OboSessionService:
    """Service interface exposing OBO-enabled endpoints to get user session information.
    """

    def __init__(self, session_api: SessionApi,
                 auth_session: AuthSession,
                 retry_config: BdkRetryConfig):
        self._session_api = session_api
        self._auth_session = auth_session
        self._retry_config = retry_config

    @retry
    async def get_session(
            self
    ) -> UserV2:
        """
        Retrieves the {@link UserV2} session from the pod using an {@link AuthSession} holder.

        Returns: User session info
        """
        params = {
            'session_token': await self._auth_session.session_token
        }
        return await self._session_api.v2_sessioninfo_get(**params)


class SessionService(OboSessionService):
    """Service class for exposing endpoints to get user session information.
    """



from symphony.bdk.core.activity.command import CommandContext, SlashCommandActivity
from symphony.bdk.core.symphony_bdk import SymphonyBdk


class HelpCommand(SlashCommandActivity):
    """The help command is a particular CommandActivity which returns the list of all commands available for the
    specific bot
    The Slash command is triggered if we receive a MESSAGESENT event with message text:
      - "@{bot_display name} /help"
    """

    def __init__(self, bdk: SymphonyBdk):
        super().__init__("/help", True, None, "List available commands")
        self._bdk = bdk

    async def on_activity(self, context: CommandContext):
        activity_list = self._bdk.activities().activity_list
        help_message = ""
        for act in activity_list:
            help_message += "<li>" + act.name + " - " + act.build_command_description() + "</li>"
        await self._bdk.messages().send_message(context.stream_id,
                                                "<messageML><ul>" + help_message + "</ul></messageML>")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['symphony',
 'symphony.bdk',
 'symphony.bdk.core',
 'symphony.bdk.core.activity',
 'symphony.bdk.core.auth',
 'symphony.bdk.core.client',
 'symphony.bdk.core.config',
 'symphony.bdk.core.config.model',
 'symphony.bdk.core.retry',
 'symphony.bdk.core.service',
 'symphony.bdk.core.service.application',
 'symphony.bdk.core.service.connection',
 'symphony.bdk.core.service.connection.model',
 'symphony.bdk.core.service.datafeed',
 'symphony.bdk.core.service.health',
 'symphony.bdk.core.service.message',
 'symphony.bdk.core.service.presence',
 'symphony.bdk.core.service.session',
 'symphony.bdk.core.service.signal',
 'symphony.bdk.core.service.stream',
 'symphony.bdk.core.service.user',
 'symphony.bdk.core.service.user.model',
 'symphony.bdk.gen',
 'symphony.bdk.gen.agent_api',
 'symphony.bdk.gen.agent_model',
 'symphony.bdk.gen.auth_api',
 'symphony.bdk.gen.auth_model',
 'symphony.bdk.gen.login_api',
 'symphony.bdk.gen.login_model',
 'symphony.bdk.gen.pod_api',
 'symphony.bdk.gen.pod_model']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.0.1,<3.0.0',
 'aiohttp>=3.7.3,<4.0.0',
 'cryptography>=3.4.6,<4.0.0',
 'defusedxml>=0.7.1,<0.8.0',
 'nulltype>=2.3.1,<3.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'tenacity>=7.0.0,<8.0.0',
 'urllib3>=1.26.2,<2.0.0']

setup_kwargs = {
    'name': 'sym-api-client-python',
    'version': '2.0b5',
    'description': 'Symphony Bot Development Kit for Python',
    'long_description': '[![License](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/licenses/MIT)\n[![Python](https://img.shields.io/badge/python-3.8%20%7C%203.9-blue)](https://www.python.org/downloads/release/python-3)\n[![Pypi](https://img.shields.io/badge/pypi-2.0b0-green)](https://pypi.org/project/sym-api-client-python/2.0b0/)\n![GitHub Workflow Status (branch)](https://img.shields.io/github/workflow/status/SymphonyPlatformSolutions/symphony-api-client-python/build/2.0)\n\n# Symphony BDK for Python\n\nThis is the Symphony BDK for Python to help develop bots and interact with the [Symphony REST APIs](https://developers.symphony.com/restapi/reference).\n\nLegacy Python BDK is located in the master branch of this repo.\n\n## Installation and getting started\nThe [reference documentation](https://symphonyplatformsolutions.github.io/symphony-api-client-python/) includes detailed\ninstallation instructions as well as a comprehensive\n[getting started](https://symphonyplatformsolutions.github.io/symphony-api-client-python/markdown/getting_started.html)\nguide.\n\n## Build from source\n\nThe Symphony BDK uses and requires Python 3.8 or higher. Be sure you have it installed before going further.\n\nWe use [poetry](https://python-poetry.org/) in order to manage dependencies, build, run tests and publish.\nTo install poetry, follow instructions [here](https://python-poetry.org/docs/#installation).\n\nOn the first time, run `poetry install`. Then run `poetry build` to build the sdist and wheel packages.\nTo run the tests, use `poetry run pytest`.\n\nIt is possible to run pylint scan locally (on a specific file or package) executing the following command:\n`poetry run pylint <module_name>`.\n\nTo generate locally the Sphinx documentation, run: `cd docsrc && make html`.\n\n## Contributing\n\nIf you want to contribute, please check the [contributing guidelines](CONTRIBUTING.md).',
    'author': 'Symphony Platform Solutions',
    'author_email': 'platformsolutions@symphony.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
    Login API

    For bots and other on-premise processes to authenticate. Once authenticated, the bot will be able to use the methods described in serviceAPI.yaml and agentAPI.yaml.  Authentication requests will expect the user to pass a token containing user identification information and signed by the user's private key  There will be two implementations of this API, one on your Pod and one on the Key Manager. In order to fully authenticate, an API client will have to call both of these implementations and pass both of the session tokens returned as headers in all subsequent requests to the Symphony API.   # noqa: E501

    The version of the OpenAPI document: 20.10.0
    Generated by: https://openapi-generator.tech
"""


import io
import json
import re
import ssl
from urllib.parse import urlencode

import aiohttp

from symphony.bdk.gen.exceptions import ApiException, UnauthorizedException, ForbiddenException, NotFoundException, ServiceException, ApiValueError


class RESTResponse(io.IOBase):

    def __init__(self, resp, data):
        self.aiohttp_response = resp
        self.status = resp.status
        self.reason = resp.reason
        self.data = data

    def getheaders(self):
        """Returns a CIMultiDictProxy of the response headers."""
        return self.aiohttp_response.headers

    def getheader(self, name, default=None):
        """Returns a given response header."""
        return self.aiohttp_response.headers.get(name, default)


class RESTClientObject(object):

    def __init__(self, configuration, pools_size=4, maxsize=None):

        # maxsize is number of requests to host that are allowed in parallel
        if maxsize is None:
            maxsize = configuration.connection_pool_maxsize

        ssl_context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH, cafile=configuration.ssl_ca_cert)
        ssl_context.load_default_certs()
        ssl_context.verify_mode = ssl.CERT_REQUIRED

        if configuration.cert_file:
            ssl_context.load_cert_chain(
                configuration.cert_file, keyfile=configuration.key_file
            )

        if not configuration.verify_ssl:
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE

        connector = aiohttp.TCPConnector(
            limit=maxsize,
            ssl=ssl_context
        )

        self.proxy = configuration.proxy
        self.proxy_headers = configuration.proxy_headers

        # https pool manager
        self.pool_manager = aiohttp.ClientSession(
            connector=connector
        )

    async def close(self):
        await self.pool_manager.close()

    async def request(self, method, url, query_params=None, headers=None,
                      body=None, post_params=None, _preload_content=True,
                      _request_timeout=None):
        """Execute request

        :param method: http request method
        :param url: http request url
        :param query_params: query parameters in the url
        :param headers: http request headers
        :param body: request json body, for `application/json`
        :param post_params: request post parameters,
                            `application/x-www-form-urlencoded`
                            and `multipart/form-data`
        :param _preload_content: this is a non-applicable field for
                                 the AiohttpClient.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        """
        method = method.upper()
        assert method in ['GET', 'HEAD', 'DELETE', 'POST', 'PUT',
                          'PATCH', 'OPTIONS']

        if post_params and body:
            raise ApiValueError(
                "body parameter cannot be used with post_params parameter."
            )

        post_params = post_params or {}
        headers = headers or {}
        timeout = _request_timeout or 5 * 60

        if 'Content-Type' not in headers:
            headers['Content-Type'] = 'application/json'

        args = {
            "method": method,
            "url": url,
            "timeout": timeout,
            "headers": headers
        }

        if self.proxy:
            args["proxy"] = self.proxy
        if self.proxy_headers:
            args["proxy_headers"] = self.proxy_headers

        if query_params:
            args["url"] += '?' + urlencode(query_params)

        # For `POST`, `PUT`, `PATCH`, `OPTIONS`, `DELETE`
        if method in ['POST', 'PUT', 'PATCH', 'OPTIONS', 'DELETE']:
            if re.search('json', headers['Content-Type'], re.IGNORECASE):
                if body is not None:
                    body = json.dumps(body)
                args["data"] = body
            elif headers['Content-Type'] == 'application/x-www-form-urlencoded':  # noqa: E501
                args["data"] = aiohttp.FormData(post_params)
            elif headers['Content-Type'] == 'multipart/form-data':
                # must del headers['Content-Type'], or the correct
                # Content-Type which generated by aiohttp
                del headers['Content-Type']
                data = aiohttp.FormData()
                for param in post_params:
                    k, v = param
                    if isinstance(v, tuple) and len(v) == 3:
                        data.add_field(k,
                                       value=v[1],
                                       filename=v[0],
                                       content_type=v[2])
                    elif isinstance(v, tuple) and len(v) == 2:
                        data.add_field(k, v[0], content_type=v[1])
                    else:
                        data.add_field(k, v)
                args["data"] = data

            # Pass a `bytes` parameter directly in the body to support
            # other content types than Json when `body` argument is provided
            # in serialized form
            elif isinstance(body, bytes):
                args["data"] = body
            else:
                # Cannot generate the request from given parameters
                msg = """Cannot prepare a request message for provided
                         arguments. Please check that your arguments match
                         declared content type."""
                raise ApiException(status=0, reason=msg)

        r = await self.pool_manager.request(**args)
        if _preload_content:

            data = await r.read()
            r = RESTResponse(r, data)

            if not 200 <= r.status <= 299:
                if r.status == 401:
                    raise UnauthorizedException(http_resp=r)

                if r.status == 403:
                    raise ForbiddenException(http_resp=r)

                if r.status == 404:
                    raise NotFoundException(http_resp=r)

                if 500 <= r.status <= 599:
                    raise ServiceException(http_resp=r)

                raise ApiException(http_resp=r)

        return r

    async def GET(self, url, headers=None, query_params=None,
                  _preload_content=True, _request_timeout=None):
        return await self.request("GET", url,
                                   headers=headers,
                                   _preload_content=_preload_content,
                                   _request_timeout=_request_timeout,
                                   query_params=query_params)

    async def HEAD(self, url, headers=None, query_params=None,
                   _preload_content=True, _request_timeout=None):
        return await self.request("HEAD", url,
                                   headers=headers,
                                   _preload_content=_preload_content,
                                   _request_timeout=_request_timeout,
                                   query_params=query_params)

    async def OPTIONS(self, url, headers=None, query_params=None,
                      post_params=None, body=None, _preload_content=True,
                      _request_timeout=None):
        return await self.request("OPTIONS", url,
                                   headers=headers,
                                   query_params=query_params,
                                   post_params=post_params,
                                   _preload_content=_preload_content,
                                   _request_timeout=_request_timeout,
                                   body=body)

    async def DELETE(self, url, headers=None, query_params=None, body=None,
                     _preload_content=True, _request_timeout=None):
        return await self.request("DELETE", url,
                                   headers=headers,
                                   query_params=query_params,
                                   _preload_content=_preload_content,
                                   _request_timeout=_request_timeout,
                                   body=body)

    async def POST(self, url, headers=None, query_params=None,
                   post_params=None, body=None, _preload_content=True,
                   _request_timeout=None):
        return await self.request("POST", url,
                                   headers=headers,
                                   query_params=query_params,
                                   post_params=post_params,
                                   _preload_content=_preload_content,
                                   _request_timeout=_request_timeout,
                                   body=body)

    async def PUT(self, url, headers=None, query_params=None, post_params=None,
                  body=None, _preload_content=True, _request_timeout=None):
        return await self.request("PUT", url,
                                   headers=headers,
                                   query_params=query_params,
                                   post_params=post_params,
                                   _preload_content=_preload_content,
                                   _request_timeout=_request_timeout,
                                   body=body)

    async def PATCH(self, url, headers=None, query_params=None,
                    post_params=None, body=None, _preload_content=True,
                    _request_timeout=None):
        return await self.request("PATCH", url,
                                   headers=headers,
                                   query_params=query_params,
                                   post_params=post_params,
                                   _preload_content=_preload_content,
                                   _request_timeout=_request_timeout,
                                   body=body)



"""
    Agent API

    This document refers to Symphony API calls to send and receive messages and content. They need the on-premise Agent installed to perform decryption/encryption of content.  - sessionToken and keyManagerToken can be obtained by calling the authenticationAPI on the symphony back end and the key manager respectively. Refer to the methods described in authenticatorAPI.yaml. - Actions are defined to be atomic, ie will succeed in their entirety or fail and have changed nothing. - If it returns a 40X status then it will have sent no message to any stream even if a request to aome subset of the requested streams would have succeeded. - If this contract cannot be met for any reason then this is an error and the response code will be 50X. - MessageML is a markup language for messages. See reference here: https://rest-api.symphony.com/docs/messagemlv2 - **Real Time Events**: The following events are returned when reading from a real time messages and events stream (\"datafeed\"). These events will be returned for datafeeds created with the v5 endpoints. To know more about the endpoints, refer to Create Messages/Events Stream and Read Messages/Events Stream. Unless otherwise specified, all events were added in 1.46.   # noqa: E501

    The version of the OpenAPI document: 20.12.0
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from symphony.bdk.gen.api_client import ApiClient, Endpoint as _Endpoint
from symphony.bdk.gen.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from symphony.bdk.gen.agent_model.error import Error
from symphony.bdk.gen.agent_model.v1_audit_trail_initiator_list import V1AuditTrailInitiatorList


class AuditTrailApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __v1_audittrail_privilegeduser_get(
            self,
            session_token,
            key_manager_token,
            start_timestamp,
            **kwargs
        ):
            """Get a list of  actions performed by a privileged account acting as privileged user given a period of time.  # noqa: E501

            Get a list of actions performed by a privileged account acting as privileged user given a period of time.  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v1_audittrail_privilegeduser_get(session_token, key_manager_token, start_timestamp, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                key_manager_token (str): Key Manager authentication token.
                start_timestamp (int): Start timestamp in unix timestamp in millseconds. 

            Keyword Args:
                end_timestamp (int): End timestamp in unix timestamp in millseconds. If not specified, it assumes to be current time. . [optional]
                before (str): Return results from an opaque “before” cursor value as presented via a response cursor. . [optional]
                after (str): Return results from an opaque “after” cursor value as presented via a response cursor. . [optional]
                limit (int): Max No. of violations to return. If no value is provided, 50 is the default. Some maximums for limit may be enforced for performance reasons. The maximum supported value is 500. . [optional]
                initiator_id (int): If present, only the initiator with this initiator <user id> will be returned. . [optional]
                role (str): If present, only the audit trail initiated by s user with privileged role acting as privileged user will be returned. Privileged eliglible roles: User Provisioning (USER_PROVISIONING), Content Management (CONTENT_MANAGEMENT), Expression Filter Policy Management (EF_POLICY_MANAGEMENT), SCO (SUPER_COMPLIANCE_OFFICER), CO (COMPLIANCE_OFFICER), Super admin (SUPER_ADMINISTRATOR), Admin (ADMINISTRATOR), L1 (L1_SUPPORT), L2 (L2_SUPPORT), Scope Manager (SCOPE_MANAGEMENT) . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                V1AuditTrailInitiatorList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['key_manager_token'] = \
                key_manager_token
            kwargs['start_timestamp'] = \
                start_timestamp
            return self.call_with_http_info(**kwargs)

        self.v1_audittrail_privilegeduser_get = _Endpoint(
            settings={
                'response_type': (V1AuditTrailInitiatorList,),
                'auth': [],
                'endpoint_path': '/v1/audittrail/privilegeduser',
                'operation_id': 'v1_audittrail_privilegeduser_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'key_manager_token',
                    'start_timestamp',
                    'end_timestamp',
                    'before',
                    'after',
                    'limit',
                    'initiator_id',
                    'role',
                ],
                'required': [
                    'session_token',
                    'key_manager_token',
                    'start_timestamp',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'key_manager_token':
                        (str,),
                    'start_timestamp':
                        (int,),
                    'end_timestamp':
                        (int,),
                    'before':
                        (str,),
                    'after':
                        (str,),
                    'limit':
                        (int,),
                    'initiator_id':
                        (int,),
                    'role':
                        (str,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'key_manager_token': 'keyManagerToken',
                    'start_timestamp': 'startTimestamp',
                    'end_timestamp': 'endTimestamp',
                    'before': 'before',
                    'after': 'after',
                    'limit': 'limit',
                    'initiator_id': 'initiatorId',
                    'role': 'role',
                },
                'location_map': {
                    'session_token': 'header',
                    'key_manager_token': 'header',
                    'start_timestamp': 'query',
                    'end_timestamp': 'query',
                    'before': 'query',
                    'after': 'query',
                    'limit': 'query',
                    'initiator_id': 'query',
                    'role': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_audittrail_privilegeduser_get
        )



import logging
import os
from abc import ABC, abstractmethod
from pathlib import Path
from symphony.bdk.core.config.model.bdk_config import BdkConfig

logger = logging.getLogger(__name__)


class DatafeedIdRepository(ABC):
    """A repository interface for storing a datafeed id.
    By using the DatafeedLoopV1, the created datafeed id and agent base url has to be persisted on the BDK side.
    """

    @abstractmethod
    def write(self, datafeed_id: str, agent_base_path: str):
        """Persists the created datafeed id into the storage.

        :param datafeed_id: the datafeed id to be persisted
        :param agent_base_path: the agent base path (i.e. scheme, host, port, context) to be persisted
        """

    @abstractmethod
    def read(self):
        """Reads the persisted datafeed from storage.

        :return the persisted datafeed id if present in storage, an empty string otherwise
        """


class OnDiskDatafeedIdRepository(DatafeedIdRepository):
    """Implementation of DatafeedIdRepository interface for persisting a datafeed id on disk."""
    DATAFEED_ID_FILE = "datafeed.id"

    def __init__(self, config: BdkConfig):
        self.config = config
        self.datafeed_id_file_path = self._get_datafeed_id_file()

    def write(self, datafeed_id, agent_base_path=""):
        self.datafeed_id_file_path.write_text(f"{datafeed_id}@{agent_base_path}")

    def read(self) -> str:
        logger.debug("Retrieving datafeed id from file %s", self.datafeed_id_file_path.absolute())
        if not os.path.exists(self.datafeed_id_file_path):
            logger.debug("Could not retrieve datafeed id from file %s: file not found",
                          self.datafeed_id_file_path.absolute())
            return None

        return self._read_in_file()

    def _read_in_file(self):
        with open(self.datafeed_id_file_path, "r") as datafeed_id_file:
            line = datafeed_id_file.readline()
            return self._read_in_line(line)

    def _read_in_line(self, line):
        index = line.find("@")
        if index == -1:
            logger.debug("Could not retrieve datafeed id from file %s: file without datafeed id information",
                          self.datafeed_id_file_path.absolute())
            return None
        datafeed_id = line[0:index]
        logger.debug("Retrieved datafeed id: %s", datafeed_id)
        return datafeed_id

    def _get_datafeed_id_file(self) -> Path:
        filepath = self.config.datafeed.get_id_file_path()
        if filepath.is_dir():
            filepath = filepath / OnDiskDatafeedIdRepository.DATAFEED_ID_FILE
        return filepath



"""This module gathers all base classes related to the datafeed loop and real time events.
"""
import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio import Task
from contextvars import ContextVar
from enum import Enum
from typing import List

from symphony.bdk.core.auth.auth_session import AuthSession
from symphony.bdk.core.config.model.bdk_config import BdkConfig
from symphony.bdk.core.service.datafeed.real_time_event_listener import RealTimeEventListener
from symphony.bdk.core.service.session.session_service import SessionService
from symphony.bdk.gen.agent_api.datafeed_api import DatafeedApi
from symphony.bdk.gen.agent_model.v4_event import V4Event

logger = logging.getLogger(__name__)

event_listener_context = ContextVar("event_listener_context", default="main-task")


class DatafeedVersion(Enum):
    """Enum of all possible datafeed versions.
    """
    V1 = "v1"
    V2 = "v2"


class RealTimeEvent(Enum):
    """This enum lists all possible
    `Real Time Events <https://docs.developers.symphony.com/building-bots-on-symphony/datafeed/real-time-events>`_
    received from the datafeed loop.

    Enum field name is the maps the exact value of field type in the datafeed payload.
    First element in enum value corresponds to the listener method who should be called when given event is received.
    Second element in enum value corresponds to the field name of event in the received payload.
    """
    MESSAGESENT = ("on_message_sent", "message_sent")
    SHAREDPOST = ("on_shared_post", "shared_post")
    INSTANTMESSAGECREATED = ("on_instant_message_created", "instant_message_created")
    ROOMCREATED = ("on_room_created", "room_created")
    ROOMUPDATED = ("on_room_updated", "room_updated")
    ROOMDEACTIVATED = ("on_room_deactivated", "room_deactivated")
    ROOMREACTIVATED = ("on_room_reactivated", "room_reactivated")
    USERREQUESTEDTOJOINROOM = ("on_user_requested_to_join_room", "user_requested_to_join_room")
    USERJOINEDROOM = ("on_user_joined_room", "user_joined_room")
    USERLEFTROOM = ("on_user_left_room", "user_left_room")
    ROOMMEMBERPROMOTEDTOOWNER = ("on_room_member_promoted_to_owner", "room_member_promoted_to_owner")
    ROOMMEMBERDEMOTEDFROMOWNER = ("on_room_demoted_from_owner", "room_member_demoted_from_owner")
    CONNECTIONREQUESTED = ("on_connection_requested", "connection_requested")
    CONNECTIONACCEPTED = ("on_connection_accepted", "connection_accepted")
    SYMPHONYELEMENTSACTION = ("on_symphony_elements_action", "symphony_elements_action")
    MESSAGESUPPRESSED = ("on_message_suppressed", "message_suppressed")


def _set_context_var(current_task, event, listener):
    event_id = getattr(event, "id", "None")
    event_listener_context.set(f"{current_task.get_name()}/{event_id}/{id(listener)}")


class AbstractDatafeedLoop(ABC):
    """Base class for implementing the datafeed services.

    A datafeed services can help a bot subscribe or unsubscribe to a RealTimeEventListener and handle the received
    event by the subscribed listeners.
    """

    def __init__(self, datafeed_api: DatafeedApi, session_service: SessionService, auth_session: AuthSession,
                 config: BdkConfig):
        """

        :param datafeed_api: The file location of the spreadsheet
        :type auth_session: the AuthSession instance used to get session and key manager tokens
        :param config: the bot configuration
        """
        self._datafeed_api = datafeed_api
        self._session_service = session_service
        self._listeners = []
        self._auth_session = auth_session
        self._api_client = datafeed_api.api_client
        self._running = False
        self._hard_kill = False
        self._timeout = None
        self._tasks = []
        self._retry_config = config.datafeed.retry
        self._bot_info = None

    async def start(self):
        """Start the datafeed event service

        :return: None
        """
        logger.debug("Starting datafeed loop")
        self._bot_info = await self._session_service.get_session()

        await self._prepare_datafeed()
        try:
            await self._run_loop()
        finally:
            logger.debug("Stopping datafeed loop")
            await self._stop_listener_tasks()

    async def stop(self, hard_kill: bool = False, timeout: float = None):
        """Stop the datafeed event service

        :param hard_kill: if set to True, tasks running listener methods will be cancelled immediately. Otherwise, tasks
          will be awaited until completion.
        :param timeout: timeout in seconds to wait for tasks completion when loop stops.
          None means wait until completion. Ignored if hard_kill set to True.
        :return: None
        """
        self._hard_kill = hard_kill
        self._timeout = timeout
        self._running = False

    def subscribe(self, listener: RealTimeEventListener):
        """Subscribes a new listener to the datafeed loop instance.

        :param listener: the RealTimeEventListener to be added.
        """
        self._listeners.append(listener)

    def unsubscribe(self, listener: RealTimeEventListener):
        """Removes a given listener from the datafeed loop instance.

        :param listener: the RealTimeEventListener to be removed.
        """
        self._listeners.remove(listener)

    @abstractmethod
    async def recreate_datafeed(self):
        """Method called when datafeed is stale and needs to be recreated (i.e. :py:meth:`read_datafeed` raises an
        ApiException with status 400)

        :return: None
        """

    @abstractmethod
    async def _prepare_datafeed(self):
        """Method called when :py:meth:`start` is called and before datafeed loop is actually running

        :return: None
        """

    async def _run_loop(self):
        self._running = True
        while self._running:
            await self._run_loop_iteration()

    @abstractmethod
    async def _run_loop_iteration(self):
        """Method called while the datafeed loop is running (.i.e stop() is called).
        This should include the logic to read events from the datafeed and dispatch them to the listeners.

        :return: None
        """

    async def _stop_listener_tasks(self):
        if self._hard_kill:
            await self._cancel_tasks()
        else:
            await self._wait_for_completion_or_timeout()

    async def _cancel_tasks(self):
        logger.debug("Cancelling %s listener tasks", len(self._tasks))
        for task in self._tasks:
            task.cancel()
        await asyncio.gather(*self._tasks, return_exceptions=True)

    async def _wait_for_completion_or_timeout(self):
        logger.debug("Waiting for %s listener tasks to finish", len(self._tasks))
        try:
            await asyncio.wait_for(asyncio.gather(*self._tasks), timeout=self._timeout)
        except asyncio.TimeoutError:
            logger.debug("Task completion timed out")

    async def _run_listener_tasks(self, events: List[V4Event]) -> List[Task]:
        tasks = await self._create_listener_tasks(events)
        if tasks:
            done, _ = await asyncio.wait(tasks)
            return done
        return []

    async def _create_listener_tasks(self, events: List[V4Event]) -> List[Task]:
        tasks = []
        sanitized_events = filter(lambda e: e is not None, events) if events else []

        for event in sanitized_events:
            for listener in self._listeners:
                if await listener.is_accepting_event(event, self._bot_info):
                    task = asyncio.create_task(self._dispatch_to_listener_method(listener, event))
                    tasks.append(task)

        return tasks

    async def _dispatch_to_listener_method(self, listener: RealTimeEventListener, event: V4Event):
        current_task = asyncio.current_task()
        _set_context_var(current_task, event, listener)

        self._tasks.append(current_task)
        try:
            await self._run_listener_method(listener, event)
        finally:
            self._tasks.remove(current_task)

    @staticmethod
    async def _run_listener_method(listener: RealTimeEventListener, event: V4Event):
        try:
            listener_method_name, payload_field_name = RealTimeEvent[event.type].value
        except KeyError:
            logger.info("Received event with an unknown type: %s", event.type)
            return

        listener_method = getattr(listener, listener_method_name)
        event_field = getattr(event.payload, payload_field_name)

        await listener_method(event.initiator, event_field)



"""Module which handles the storage of valid extension app tokens.
"""
from abc import ABC, abstractmethod


class TokensRepository(ABC):
    """Base abstract class to store and retrieve extension app tokens.
    """

    @abstractmethod
    async def save(self, app_token: str, symphony_token: str) -> None:
        """Saves a pair (app_token, symphony_token)

        :param app_token: the application token
        :param symphony_token: the Symphony token
        :return: None
        """

    @abstractmethod
    async def get(self, app_token: str) -> str:
        """Retrieves the corresponding Symphony token from a given application token.

        :param app_token: the application token
        :return: the symphony token corresponding to the app token if it exists, None otherwise
        """


class InMemoryTokensRepository(TokensRepository):
    """Class implementing an in-memory TokensRepository.
    """
    def __init__(self):
        self._tokens = {}

    async def save(self, app_token: str, symphony_token: str) -> None:
        self._tokens[app_token] = symphony_token

    async def get(self, app_token: str) -> str:
        return self._tokens.get(app_token)



"""Service factory module"""
from symphony.bdk.core.auth.auth_session import AuthSession
from symphony.bdk.core.client.api_client_factory import ApiClientFactory
from symphony.bdk.core.config.model.bdk_config import BdkConfig
from symphony.bdk.core.service.application.application_service import ApplicationService
from symphony.bdk.core.service.connection.connection_service import ConnectionService, OboConnectionService
from symphony.bdk.core.service.datafeed.abstract_datafeed_loop import DatafeedVersion, AbstractDatafeedLoop
from symphony.bdk.core.service.datafeed.datafeed_loop_v1 import DatafeedLoopV1
from symphony.bdk.core.service.datafeed.datafeed_loop_v2 import DatafeedLoopV2
from symphony.bdk.core.service.health.health_service import HealthService
from symphony.bdk.core.service.message.message_service import MessageService, OboMessageService
from symphony.bdk.core.service.message.multi_attachments_messages_api import MultiAttachmentsMessagesApi
from symphony.bdk.core.service.presence.presence_service import PresenceService, OboPresenceService
from symphony.bdk.core.service.session.session_service import SessionService
from symphony.bdk.core.service.signal.signal_service import SignalService, OboSignalService
from symphony.bdk.core.service.stream.stream_service import StreamService, OboStreamService
from symphony.bdk.core.service.user.user_service import UserService, OboUserService
from symphony.bdk.gen.agent_api.attachments_api import AttachmentsApi
from symphony.bdk.gen.agent_api.audit_trail_api import AuditTrailApi
from symphony.bdk.gen.agent_api.datafeed_api import DatafeedApi
from symphony.bdk.gen.agent_api.share_api import ShareApi
from symphony.bdk.gen.agent_api.signals_api import SignalsApi
from symphony.bdk.gen.agent_api.system_api import SystemApi as AgentSystemApi
from symphony.bdk.gen.pod_api.app_entitlement_api import AppEntitlementApi
from symphony.bdk.gen.pod_api.application_api import ApplicationApi
from symphony.bdk.gen.pod_api.connection_api import ConnectionApi
from symphony.bdk.gen.pod_api.default_api import DefaultApi
from symphony.bdk.gen.pod_api.message_api import MessageApi
from symphony.bdk.gen.pod_api.message_suppression_api import MessageSuppressionApi
from symphony.bdk.gen.pod_api.pod_api import PodApi
from symphony.bdk.gen.pod_api.presence_api import PresenceApi
from symphony.bdk.gen.pod_api.room_membership_api import RoomMembershipApi
from symphony.bdk.gen.pod_api.session_api import SessionApi
from symphony.bdk.gen.pod_api.streams_api import StreamsApi
from symphony.bdk.gen.pod_api.system_api import SystemApi as PodSystemApi
from symphony.bdk.gen.pod_api.user_api import UserApi
from symphony.bdk.gen.pod_api.users_api import UsersApi


class ServiceFactory:
    """Factory responsible for creating BDK service instances for Symphony Bdk entry point:

    * User Service
    * Message Service
    * Connection Service
    * Stream Service
    * Datafeed Loop
    * ...
    """

    def __init__(
            self,
            api_client_factory: ApiClientFactory,
            auth_session: AuthSession,
            config: BdkConfig
    ):
        self._pod_client = api_client_factory.get_pod_client()
        self._agent_client = api_client_factory.get_agent_client()
        self._auth_session = auth_session
        self._config = config
        self._session_service = SessionService(SessionApi(self._pod_client), self._auth_session, self._config.retry)

    def get_user_service(self) -> UserService:
        """Returns a fully initialized UserService

        :return: a new UserService instance.
        """
        return UserService(
            UserApi(self._pod_client),
            UsersApi(self._pod_client),
            AuditTrailApi(self._agent_client),
            PodSystemApi(self._pod_client),
            self._auth_session,
            self._config.retry
        )

    def get_message_service(self) -> MessageService:
        """Returns a fully initialized MessageService

        :return: a new MessageService instance.
        """
        return MessageService(
            MultiAttachmentsMessagesApi(self._agent_client),
            MessageApi(self._pod_client),
            MessageSuppressionApi(self._pod_client),
            StreamsApi(self._pod_client),
            PodApi(self._pod_client),
            AttachmentsApi(self._agent_client),
            DefaultApi(self._pod_client),
            self._auth_session,
            self._config.retry
        )

    def get_connection_service(self) -> ConnectionService:
        """Returns a fully initialized ConnectionService

        :return: a new ConnectionService instance.
        """
        return ConnectionService(
            ConnectionApi(self._pod_client),
            self._auth_session,
            self._config.retry
        )

    def get_stream_service(self) -> StreamService:
        """Returns a fully initialized StreamService

        :return: a new StreamService instance
        """
        return StreamService(
            StreamsApi(self._pod_client),
            RoomMembershipApi(self._pod_client),
            ShareApi(self._agent_client),
            self._auth_session,
            self._config.retry)

    def get_application_service(self) -> ApplicationService:
        """Returns a fully initialized ApplicationService

        :return: a new ApplicationService instance
        """
        return ApplicationService(
            ApplicationApi(self._pod_client),
            AppEntitlementApi(self._pod_client),
            self._auth_session,
            self._config.retry
        )

    def get_signal_service(self) -> SignalService:
        """Returns a fully initialized SignalService

        :return: a new SignalService instance
        """
        return SignalService(
            SignalsApi(self._agent_client),
            self._auth_session,
            self._config.retry
        )

    def get_session_service(self) -> SessionService:
        """Returns a fully initialized SessionService

        :return: a new SessionService instance
        """
        return self._session_service

    def get_datafeed_loop(self) -> AbstractDatafeedLoop:
        """Returns a fully initialized DatafeedLoop

        :return: a new DatafeedLoop instance.
        """
        df_version = self._config.datafeed.version
        if df_version.lower() == DatafeedVersion.V2.value.lower():
            return DatafeedLoopV2(
                DatafeedApi(self._agent_client),
                self._session_service,
                self._auth_session,
                self._config
            )
        return DatafeedLoopV1(
            DatafeedApi(self._agent_client),
            self._session_service,
            self._auth_session,
            self._config
        )

    def get_health_service(self) -> HealthService:
        """Returns a fully initialized HealthService

        :return: a new HealthService instance
        """
        return HealthService(
            AgentSystemApi(self._agent_client),
            SignalsApi(self._agent_client),
            self._config.retry
        )

    def get_presence_service(self) -> PresenceService:
        """Returns a fully initialized PresenceService

        :return: a new PresenceService instance
        """
        return PresenceService(
            PresenceApi(self._pod_client),
            self._auth_session,
            self._config.retry
        )


class OboServiceFactory:
    """Factory responsible for creating BDK service instances for OBO-enabled endpoints only:

    * User Service
    * Message Service
    * Connection Service
    * Stream Service

    See: `OBO-enabled endpoints <https://developers.symphony.com/restapi/reference#obo-enabled-endpoints>`_
    """

    def __init__(
            self,
            api_client_factory: ApiClientFactory,
            auth_session: AuthSession,
            config: BdkConfig
    ):
        self._pod_client = api_client_factory.get_pod_client()
        self._agent_client = api_client_factory.get_agent_client()
        self._auth_session = auth_session
        self._config = config

    def get_user_service(self) -> OboUserService:
        """Returns a fully initialized OboUserService

        :return: a new OboUserService instance.
        """
        return OboUserService(
            UserApi(self._pod_client),
            UsersApi(self._pod_client),
            self._auth_session,
            self._config.retry
        )

    def get_message_service(self) -> OboMessageService:
        """Returns a fully initialized OboMessageService

        :return: a new OboMessageService instance.
        """
        return OboMessageService(
            MultiAttachmentsMessagesApi(self._agent_client),
            MessageSuppressionApi(self._pod_client),
            self._auth_session,
            self._config.retry
        )

    def get_connection_service(self) -> OboConnectionService:
        """Returns a fully initialized OboConnectionService

        :return: a new OboConnectionService instance.
        """
        return OboConnectionService(
            ConnectionApi(self._pod_client),
            self._auth_session,
            self._config.retry
        )

    def get_stream_service(self) -> OboStreamService:
        """Returns a fully initialized OboStreamService

        :return: a new OboStreamService instance
        """
        return OboStreamService(
            StreamsApi(self._pod_client),
            RoomMembershipApi(self._pod_client),
            ShareApi(self._agent_client),
            self._auth_session,
            self._config.retry)

    def get_presence_service(self) -> OboPresenceService:
        """Returns a fully initialized OboPresenceService

        :return: a new OboPresenceService instance
        """
        return OboPresenceService(
            PresenceApi(self._pod_client),
            self._auth_session,
            self._config.retry
        )

    def get_signal_service(self) -> OboSignalService:
        """Returns a fully initialized OboSignalService

        :return: a new OboSignalService instance
        """
        return OboSignalService(
            SignalsApi(self._agent_client),
            self._auth_session,
            self._config.retry
        )



from symphony.bdk.core.config.model.bdk_authentication_config import BdkAuthenticationConfig


class BdkAppConfig(BdkAuthenticationConfig):
    """Class containing the extension app configuration
    """

    def __init__(self, config):
        if config is not None:
            self.app_id = config.get("appId")
            super().__init__(private_key_config=config.get("privateKey"), certificate_config=config.get("certificate"))
        else:
            super().__init__()



from symphony.bdk.gen.agent_model.v4_event import V4Event
from symphony.bdk.gen.agent_model.v4_message_sent import V4MessageSent
from symphony.bdk.gen.agent_model.v4_initiator import V4Initiator
from symphony.bdk.gen.agent_model.v4_shared_post import V4SharedPost
from symphony.bdk.gen.agent_model.v4_instant_message_created import V4InstantMessageCreated
from symphony.bdk.gen.agent_model.v4_room_created import V4RoomCreated
from symphony.bdk.gen.agent_model.v4_room_updated import V4RoomUpdated
from symphony.bdk.gen.agent_model.v4_room_deactivated import V4RoomDeactivated
from symphony.bdk.gen.agent_model.v4_room_reactivated import V4RoomReactivated
from symphony.bdk.gen.agent_model.v4_user_requested_to_join_room import V4UserRequestedToJoinRoom
from symphony.bdk.gen.agent_model.v4_user_joined_room import V4UserJoinedRoom
from symphony.bdk.gen.agent_model.v4_user_left_room import V4UserLeftRoom
from symphony.bdk.gen.agent_model.v4_room_member_promoted_to_owner import V4RoomMemberPromotedToOwner
from symphony.bdk.gen.agent_model.v4_room_member_demoted_from_owner import V4RoomMemberDemotedFromOwner
from symphony.bdk.gen.agent_model.v4_connection_requested import V4ConnectionRequested
from symphony.bdk.gen.agent_model.v4_connection_accepted import V4ConnectionAccepted
from symphony.bdk.gen.agent_model.v4_message_suppressed import V4MessageSuppressed
from symphony.bdk.gen.agent_model.v4_symphony_elements_action import V4SymphonyElementsAction
from symphony.bdk.gen.pod_model.user_v2 import UserV2


class RealTimeEventListener:
    """Interface for a callback to be invoked when a RealTimeEvent is received from the datafeed
    `real-time-events <https://developers.symphony.com/restapi/docs/real-time-events>_`
    """

    @staticmethod
    async def is_accepting_event(event: V4Event, bot_info: UserV2) -> bool:
        """Checks if the event is accepted to be handled.

        By default, all the events that is created by the bot itself will not be accepted to be handled by the
        listener. If you want to handle the self-created events or you want to apply your own filters for the events,
        you should override this method.

        :param event: Event to be verified.
        :param bot_info: the UserV2 object containing the bot service account information
        :return: True if the event is accepted, False otherwise
        """
        try:
            return event.initiator.user.user_id != bot_info.id
        except AttributeError:
            return False

    async def on_message_sent(self, initiator: V4Initiator, event: V4MessageSent):
        """ Called when a MESSAGESENT event is received.

        :param initiator: Event initiator.
        :param event: Message sent payload.
        """

    async def on_shared_post(self, initiator: V4Initiator, event: V4SharedPost):
        """ Called when an INSTANTMESSAGECREATED event is received.

        :param initiator: Event initiator.
        :param event:  Shared post payload.
        """

    async def on_instant_message_created(self, initiator: V4Initiator, event: V4InstantMessageCreated):
        """ Called when a ROOMCREATED event is received.

        :param initiator: Event initiator.
        :param event: Instant Message Created payload.
        """

    async def on_room_created(self, initiator: V4Initiator, event: V4RoomCreated):
        """

        :param initiator: Event initiator.
        :param event: Room Created payload.
        """

    async def on_room_updated(self, initiator: V4Initiator, event: V4RoomUpdated):
        """

        :param initiator: Event initiator.
        :param event: Room Updated payload.
        """

    async def on_room_deactivated(self, initiator: V4Initiator, event: V4RoomDeactivated):
        """

        :param initiator: Event initiator.
        :param event: Room Deactivated payload.
        """

    async def on_room_reactivated(self, initiator: V4Initiator, event: V4RoomReactivated):
        """

        :param initiator: Event initiator.
        :param event: Room Reactivated payload.
        """

    async def on_user_requested_to_join_room(self, initiator: V4Initiator, event: V4UserRequestedToJoinRoom):
        """

        :param initiator: Event initiator.
        :param event: User Requested To Join Room payload.
        """

    async def on_user_joined_room(self, initiator: V4Initiator, event: V4UserJoinedRoom):
        """

        :param initiator: Event initiator.
        :param event: User Joined Room payload.
        """

    async def on_user_left_room(self, initiator: V4Initiator, event: V4UserLeftRoom):
        """

        :param initiator: Event initiator.
        :param event: User Left Room payload.
        """

    async def on_room_member_promoted_to_owner(self, initiator: V4Initiator, event: V4RoomMemberPromotedToOwner):
        """

        :param initiator: Event initiator.
        :param event: Room Member Promoted To Owner payload.
        """

    async def on_room_demoted_from_owner(self, initiator: V4Initiator, event: V4RoomMemberDemotedFromOwner):
        """

        :param initiator: Event initiator.
        :param event: Room Member Demoted From Owner payload.
        """

    async def on_connection_requested(self, initiator: V4Initiator, event: V4ConnectionRequested):
        """

        :param initiator: Event initiator.
        :param event: Connection Requested payload.
        """

    async def on_connection_accepted(self, initiator: V4Initiator, event: V4ConnectionAccepted):
        """

        :param initiator: Event initiator.
        :param event: Connection Accepted payload.
        """

    async def on_message_suppressed(self, initiator: V4Initiator, event: V4MessageSuppressed):
        """

        :param initiator: Event initiator.
        :param event: Message Suppressed payload.
        """

    async def on_symphony_elements_action(self, initiator: V4Initiator, event: V4SymphonyElementsAction):
        """

        :param initiator: Event initiator.
        :param event: Symphony Elements Action payload.
        """



from symphony.bdk.gen.agent_api.messages_api import MessagesApi
from symphony.bdk.gen.agent_model.v4_message import V4Message
from symphony.bdk.gen.agent_model.v4_message_blast_response import V4MessageBlastResponse
from symphony.bdk.gen.api_client import Endpoint
from symphony.bdk.gen.model_utils import (
    file_type
)


class MultiAttachmentsMessagesApi(MessagesApi):
    """Message Api inherited the MessagesApi generated by OpenAPI Generator
    which allows to send or blast a message with multiple attachments and previews
    """

    def __init__(self, api_client=None):
        super().__init__(api_client)

        def __v4_message_create_post(self, **kwargs):
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        self.v4_stream_sid_multi_attachment_message_create_post = Endpoint(
            settings={
                'response_type': (V4Message,),
                'auth': [],
                'endpoint_path': '/v4/stream/{sid}/message/create',
                'operation_id': 'v4_stream_sid_message_create_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'sid',
                    'session_token',
                    'key_manager_token',
                    'message',
                    'data',
                    'version',
                    'attachment',
                    'preview',
                ],
                'required': [
                    'sid',
                    'session_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'sid':
                        (str,),
                    'session_token':
                        (str,),
                    'key_manager_token':
                        (str,),
                    'message':
                        (str,),
                    'data':
                        (str,),
                    'version':
                        (str,),
                    'attachment':
                        ([file_type],),
                    'preview':
                        ([file_type],),
                },
                'attribute_map': {
                    'sid': 'sid',
                    'session_token': 'sessionToken',
                    'key_manager_token': 'keyManagerToken',
                    'message': 'message',
                    'data': 'data',
                    'version': 'version',
                    'attachment': 'attachment',
                    'preview': 'preview',
                },
                'location_map': {
                    'sid': 'path',
                    'session_token': 'header',
                    'key_manager_token': 'header',
                    'message': 'form',
                    'data': 'form',
                    'version': 'form',
                    'attachment': 'form',
                    'preview': 'form',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'multipart/form-data'
                ]
            },
            api_client=api_client,
            callable=__v4_message_create_post
        )

        self.v4_multi_attachment_message_blast_post = Endpoint(
            settings={
                'response_type': (V4MessageBlastResponse,),
                'auth': [],
                'endpoint_path': '/v4/message/blast',
                'operation_id': 'v4_message_blast_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'sids',
                    'key_manager_token',
                    'message',
                    'data',
                    'version',
                    'attachment',
                    'preview',
                ],
                'required': [
                    'session_token',
                    'sids',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'sids':
                        ([str],),
                    'key_manager_token':
                        (str,),
                    'message':
                        (str,),
                    'data':
                        (str,),
                    'version':
                        (str,),
                    'attachment':
                        ([file_type],),
                    'preview':
                        ([file_type],),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'sids': 'sids',
                    'key_manager_token': 'keyManagerToken',
                    'message': 'message',
                    'data': 'data',
                    'version': 'version',
                    'attachment': 'attachment',
                    'preview': 'preview',
                },
                'location_map': {
                    'session_token': 'header',
                    'sids': 'form',
                    'key_manager_token': 'header',
                    'message': 'form',
                    'data': 'form',
                    'version': 'form',
                    'attachment': 'form',
                    'preview': 'form',
                },
                'collection_format_map': {
                    'sids': 'csv',
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'multipart/form-data'
                ]
            },
            api_client=api_client,
            callable=__v4_message_create_post
        )
